import functools
import tensorflow as tf
from gpflow import default_float
from gpflow.utilities import to_default_float
from gpflow.utilities.ops import square_distance, difference_matrix


def cached(variable):
    def cache(func):
        func = tf.function(func, experimental_compile=True, experimental_relax_shapes=True)

        @functools.wraps(func)
        def caching_wrapper(self, *args, **kwargs):
            if not hasattr(self, variable) or getattr(self, variable) is None:
                mat = func(self.X)
                if self._cache:
                    setattr(self, variable, mat)
            else:
                mat = getattr(self, variable)
            return mat

        return caching_wrapper

    return cache


class DistanceCache:
    def __init__(self, X: tf.Tensor, cache=True):
        self.X = X
        self._cache = cache

    @property
    @cached("_squaredEuclidean")
    def squaredEuclideanDistance(X):
        return square_distance(to_default_float(X), None)

    @property
    @cached("_sumDiff")
    def sumOfDifferences(X):
        return tf.reduce_sum(difference_matrix(to_default_float(X), None), axis=-1)



# from .version import version as __version__

from .de_test import test

from .gaussian_process import GP, SGPIPM, GPControl, fit, fit_fast, fit_detailed
from .dp_hmrf import (
    tissue_segmentation,
    TissueSegmentationParameters,
    TissueSegmentationStatus,
    TissueSegmentation,
)
from .aeh import spatial_patterns, SpatialPatternParameters, SpatialPatterns
from .svca import test_spatial_interactions, fit_spatial_interactions
from .io import read_spaceranger

import tensorflow as tf

gpus = tf.config.experimental.list_physical_devices("GPU")
if gpus:
    try:
        for gpu in gpus:
            tf.config.experimental.set_memory_growth(gpu, True)
        logical_gpus = tf.config.experimental.list_logical_devices("GPU")
    except RuntimeError as e:
        print(e)
del tf
del gpus



import os
import glob
import warnings
import json
import logging

import numpy as np
from scipy.sparse import csr_matrix
import pandas as pd
from matplotlib.image import imread

import h5py
from anndata import AnnData


def read_spaceranger(spaceranger_out_dir: str, read_images: bool = True) -> AnnData:
    """
    Read 10x SpaceRanger output.

    Args:
        spaceranger_out_dir: Path to the directory with SpaceRanger output.
        read_images: Whether to also read images into memory.

    Returns:
        An annotated data matrix.
    """
    fname = glob.glob(os.path.join(spaceranger_out_dir, "*filtered_feature_bc_matrix.h5"))
    if len(fname) == 0:
        raise FileNotFoundError(
            "filtered_feature_bc_matrix.h5 file not found in specified directory"
        )
    elif len(fname) > 1:
        warnings.warn(
            "Multiple files ending with filtered_feature_bc_matrix.h5 found in specified directory, using the first one",
            RuntimeWarning,
        )
    fname = fname[0]
    with h5py.File(fname, "r") as f:
        matrix = f["matrix"]
        sparsemat = csr_matrix(
            (matrix["data"][...], matrix["indices"][...], matrix["indptr"][...]),
            shape=matrix["shape"][...][::-1],
        )

        barcodes = matrix["barcodes"][...].astype(np.unicode)

        adata = AnnData(X=sparsemat)

        features = matrix["features"]
        adata.var_names = features["name"][...].astype(np.unicode)
        adata.var["id"] = features["id"][...].astype(np.unicode)
        for f in features["_all_tag_keys"]:
            feature = features[f][...]
            if feature.dtype.kind in ("S", "U"):
                feature = feature.astype(np.unicode)
            adata.var[f.astype(np.unicode)] = feature

    _, counts = np.unique(adata.var_names, return_counts=True)
    if np.sum(counts > 1) > 0:
        logging.warning("Duplicate gene names present. Converting to unique names.")
        adata.var_names_make_unique()

    tissue_positions = (
        pd.read_csv(
            os.path.join(spaceranger_out_dir, "spatial", "tissue_positions_list.csv"),
            names=(
                "barcode",
                "in_tissue",
                "array_row",
                "array_col",
                "pxl_col_in_fullres",
                "pxl_row_in_fullres",
            ),
        )
        .set_index("barcode")
        .loc[barcodes]
        .drop("in_tissue", axis=1)
    )
    adata.obsm["spatial"] = tissue_positions[
        ["pxl_row_in_fullres", "pxl_col_in_fullres"]
    ].to_numpy()
    adata.obs = tissue_positions.drop(["pxl_row_in_fullres", "pxl_col_in_fullres"], axis=1)

    with open(os.path.join(spaceranger_out_dir, "spatial", "scalefactors_json.json"), "r") as f:
        meta = json.load(f)
    adata.uns["spot_diameter_fullres"] = meta["spot_diameter_fullres"]
    if read_images:
        adata.uns["tissue_lowres_image"] = imread(
            os.path.join(spaceranger_out_dir, "spatial", "tissue_lowres_image.png")
        )
        adata.uns["tissue_hires_image"] = imread(
            os.path.join(spaceranger_out_dir, "spatial", "tissue_hires_image.png")
        )
        adata.uns["tissue_hires_scalef"] = meta["tissue_hires_scalef"]
        adata.uns["tissue_lowres_scalef"] = meta["tissue_lowres_scalef"]

    return adata



import scanpy as sc
import numpy as np
import matplotlib.pyplot as plt
import squidpy as sq

def smooth_pattern(idata, n_neighbors=10):
    import sklearn
    neigh = sklearn.neighbors.NearestNeighbors(n_neighbors=10)
    neigh.fit(idata.obs[['row', 'col']])
    neighbor_idx = neigh.kneighbors(idata.obs[['row', 'col']], return_distance=False)

    smoothed_pattern_score = []
    for i in neighbor_idx:
        smoothed_pattern_score.append(np.sum(idata.obsm['pattern_score'][i], axis=0))
    smoothed_pattern_score = np.array(smoothed_pattern_score)
    idata.obsm['smooth_pattern_score'] = smoothed_pattern_score

def paga(idata, label, n_neighbors=30, min_dist=1):
    import umap
    embedding = umap.UMAP(n_neighbors=n_neighbors, min_dist=min_dist,random_state=52).fit_transform(idata.obsm['smooth_pattern_score'])
    idata.obsm['X_umap'] = embedding
    sc.pp.neighbors(idata, n_neighbors=20, use_rep='X_umap')
    sc.tl.paga(idata, groups=label)
    

def paga_default(idata, label):
    import umap
    embedding = umap.UMAP(random_state=52, n_components=3).fit_transform(idata.obsm['smooth_pattern_score'])
    idata.obsm['X_umap'] = embedding
    sc.pp.neighbors(idata, use_rep='X_umap')
    sc.tl.paga(idata, groups=label)
    
def paga_spot(idata, adata, label, n_neighbors=30, min_dist=1):
    import umap
    embedding = umap.UMAP(random_state=42, min_dist=0.8, n_neighbors=20).fit_transform(idata.uns['cell_pattern'])
    adata.obsm['X_umap'] = embedding
    sc.pp.neighbors(adata, n_neighbors=20, use_rep='X_umap')   
    sc.tl.paga(adata, groups=label)
    
def pseudotime(idata, root_label, root_id=0):
    idata.uns['iroot'] = np.flatnonzero(idata.obs['label']  == root_label)[root_id]
    sc.tl.dpt(idata)
    
def projection(idata):
    import cellrank as cr
    import networkx
    networkx.from_scipy_sparse_matrix = networkx.from_scipy_sparse_array
    sq.gr.spatial_neighbors(idata, key_added='spatial')
    pt=cr.kernels.PseudotimeKernel(idata, time_key='dpt_pseudotime')
    pt.compute_transition_matrix()
    ck = cr.kernels.ConnectivityKernel(idata,conn_key='spatial_connectivities')
    ck.compute_transition_matrix()
    k=9*pt+1*ck
    k.compute_transition_matrix()
    k.compute_projection(basis="spatial")
    



import tensorflow as tf
import tensorflow_probability as tfp
import numpy as np
import matplotlib.pyplot as plt

from gpflow import Parameter
from gpflow.kernels import Stationary, Sum
from gpflow.utilities import positive
from gpflow.utilities.ops import square_distance, difference_matrix
from gpflow.utilities import to_default_float


class Spectral(Stationary):
    def __init__(self, variance=1.0, lengthscales=1, periods=1, **kwargs):
        super().__init__(
            variance=variance,
            lengthscales=Parameter(lengthscales, transform=positive(lower=to_default_float(1e-6))),
            **kwargs,
        )
        self.periods = Parameter(periods, transform=positive(lower=to_default_float(1e-6)))

        self._validate_ard_active_dims(self.periods)

    @property
    def ard(self) -> bool:
        return self.lengthscales.shapes.ndims > 0 or self.periods.shapes.ndims > 0

    def K(self, X, X2=None):
        return self.variance * self.K_novar(X, X2)

    def K_novar(self, X, X2=None):
        Xs = X / self.periods
        if X2 is not None:
            X2s = X2 / self.periods
        else:
            X2s = None
        dist = difference_matrix(Xs, X2s)
        cospart = tf.cos(2 * np.pi * tf.reduce_sum(dist, axis=-1))

        dist = square_distance(self.scale(X), self.scale(X2))
        exppart = tf.exp(-0.5 * dist)

        return cospart * exppart

    def K_diag(self, X):
        return tf.fill(tf.shape(X)[:-1], tf.squeeze(self.variance))

    def log_power_spectrum(self, s):
        if not tf.is_tensor(s):
            s = tf.convert_to_tensor(s, dtype=self.variance.dtype)
        else:
            s = tf.cast(s, dtype=self.variance.dtype)
        if s.ndim < 2:
            s = tf.expand_dims(s, 1)
        loc = tf.broadcast_to(self.periods, (s.shape[1],))
        scale_diag = tf.broadcast_to(self.lengthscales / (0.25 * np.pi**2), (s.shape[1],))
        mvd = tfp.distributions.MultivariateNormalDiag(loc=1 / loc, scale_diag=1 / scale_diag)
        return tf.math.log(tf.constant(0.5, dtype=self.variance.dtype)) + tf.reduce_logsumexp(
            [mvd.log_prob(s), mvd.log_prob(-s)], axis=0
        )


class SpectralMixture(Sum):
    def __init__(self, kernels=None, dimnames=None, **kwargs):
        if kernels is None:
            kernels = [Spectral()]
        elif isinstance(kernels, list):
            if not all([isinstance(k, Spectral) for k in kernels]):
                raise ValueError("Not all kernels are Spectral")
        else:
            kernels = [Spectral() for _ in range(kernels)]
        super().__init__(kernels)

        if dimnames is None:
            self.dimnames = ("X", "Y")
        else:
            self.dimnames = dimnames

    def K_novar(self, X, X2=None):
        return self._reduce([k.K_novar(X, X2) for k in self.kernels])

    def log_power_spectrum(self, s):
        dens = []
        for k in self.kernels:
            dens.append(k.variance * k.log_power_spectrum(s))
        return tf.reduce_logsumexp(dens, axis=0)

    def plot_power_spectrum(self, xlim: float = None, ylim: float = None, **kwargs):
        if xlim is None or ylim is None:
            lengthscales = tf.convert_to_tensor([k.lengthscales for k in self.kernels])
            if lengthscales.ndim < 2:
                lengthscales = tf.tile(tf.expand_dims(lengthscales, axis=1), (1, 2))
            periods = tf.convert_to_tensor([k.periods for k in self.kernels])
            if periods.ndim < 2:
                periods = tf.tile(tf.expand_dims(periods, axis=1), (1, 2))
            maxfreq = tf.math.argmin(periods, axis=0, output_type=tf.int32)
            maxfreq = tf.stack([maxfreq, tf.range(maxfreq.shape[0], dtype=tf.int32)], axis=1)
            limits = 1 / tf.gather_nd(periods, maxfreq)
            limits += 2 * tf.gather_nd(lengthscales, maxfreq)
        if xlim is None:
            xlim = limits[0].numpy()
        else:
            xlim = np.asarray([xlim])[0]
        if ylim is None:
            ylim = limits[1].numpy()
        else:
            ylim = np.asarray([ylim])[0]

        limtype = np.promote_types(xlim.dtype, ylim.dtype)
        xlim = xlim.astype(limtype)
        ylim = ylim.astype(limtype)

        dim = max(
            [k.lengthscales.ndim for k in self.kernels] + [k.periods.ndim for k in self.kernels]
        )
        fig, ax = plt.subplots()
        if dim < 1:
            x = tf.linspace(0, xlim, 1000)
            ps = self.log_power_spectrum(x)
            ax.plot(x, ps)
            ax.set_xlim((0, xlim.numpy()))
            ax.set_xlabel("frequency")
            ax.set_ylabel("log spectral density")
        else:
            x, y = tf.meshgrid(tf.linspace(0.0, xlim, 1000), tf.linspace(0.0, ylim, 1000))
            ps = tf.reshape(
                self.log_power_spectrum(
                    tf.stack([tf.reshape(x, (-1,)), tf.reshape(y, (-1,))], axis=1)
                ),
                x.shape,
            )
            pos = ax.contourf(
                x, y, ps, levels=tf.linspace(tf.reduce_min(ps), tf.reduce_max(ps), 100), **kwargs
            )
            ax.set_xlabel(f"{self.dimnames[0]} frequency")
            ax.set_ylabel(f"{self.dimnames[1]} frequency")
            cbar = fig.colorbar(pos)
            cbar.ax.set_ylabel("log spectral density")
        return ax

    def __iter__(self):
        self._i = 0
        return self

    def __next__(self):
        if self._i < len(self.kernels):
            k = self.kernels[self._i]
            self._i += 1
            return k
        else:
            raise StopIteration
